#!/usr/bin/env python3
"""
Port Knocking Sequence Tool
Parameters:
    - IP: Target IP address
    - ports: Comma-separated list of ports in sequence
Additional parameters in comments:
    - delay: Delay between knocks in seconds
    - protocol: TCP or UDP
"""
from scapy.all import *
import argparse
import time
from datetime import datetime
import logging
from pathlib import Path

class PortKnocker:
    def __init__(self, target_ip, port_sequence, delay=0.5, protocol='tcp'):
        self.target_ip = target_ip
        self.port_sequence = port_sequence
        self.delay = delay
        self.protocol = protocol.lower()
        self.setup_logging()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("knock_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "port_knocker.log"),
                logging.StreamHandler()
            ]
        )
        
    def create_packet(self, port):
        """Create TCP or UDP packet"""
        if self.protocol == 'tcp':
            return IP(dst=self.target_ip)/TCP(dport=port, flags="S")
        else:
            return IP(dst=self.target_ip)/UDP(dport=port)
            
    def verify_port(self, port):
        """Verify if port is open after knocking sequence"""
        try:
            # Try TCP connection
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(2)
            result = sock.connect_ex((self.target_ip, port))
            sock.close()
            return result == 0
        except:
            return False
            
    def knock(self, verify_port=None):
        """Perform port knocking sequence"""
        logging.info(f"Starting port knock sequence on {self.target_ip}")
        logging.info(f"Protocol: {self.protocol.upper()}")
        logging.info(f"Sequence: {self.port_sequence}")
        logging.info(f"Delay: {self.delay} seconds")
        
        start_time = time.time()
        success = True
        
        try:
            # Send knock sequence
            for i, port in enumerate(self.port_sequence, 1):
                packet = self.create_packet(port)
                
                # Send packet
                send(packet, verbose=0)
                logging.info(f"Knock {i}/{len(self.port_sequence)}: Port {port}")
                
                # Wait between knocks
                if i < len(self.port_sequence):
                    time.sleep(self.delay)
                    
            duration = time.time() - start_time
            logging.info(f"Knock sequence completed in {duration:.2f} seconds")
            
            # Verify target port if specified
            if verify_port:
                logging.info(f"Verifying target port {verify_port}...")
                time.sleep(1)  # Wait for port to open
                
                if self.verify_port(verify_port):
                    logging.info(f"Success: Port {verify_port} is now open")
                else:
                    logging.warning(f"Warning: Port {verify_port} appears to be closed")
                    success = False
                    
            return success
            
        except Exception as e:
            logging.error(f"Error during port knocking: {str(e)}")
            return False
            
    def scan_for_response(self, ports_to_check):
        """Scan for any response after knock sequence"""
        logging.info("Scanning for responses...")
        
        responses = []
        for port in ports_to_check:
            if self.verify_port(port):
                responses.append(port)
                logging.info(f"Found open port: {port}")
                
        return responses
        
    def perform_advanced_knock(self, verify_ports=None):
        """Perform advanced port knocking with verification"""
        logging.info("Starting advanced port knocking sequence")
        
        # Initial state check
        if verify_ports:
            logging.info("Checking initial port states...")
            initial_states = {port: self.verify_port(port) for port in verify_ports}
            
        # Perform knock sequence
        success = self.knock()
        
        if not success:
            logging.error("Knock sequence failed")
            return False
            
        # Verify results if ports specified
        if verify_ports:
            logging.info("Verifying results...")
            time.sleep(1)  # Wait for changes to take effect
            
            changes = []
            for port in verify_ports:
                new_state = self.verify_port(port)
                if new_state != initial_states[port]:
                    changes.append({
                        'port': port,
                        'initial_state': initial_states[port],
                        'final_state': new_state
                    })
                    
            if changes:
                logging.info("Detected port state changes:")
                for change in changes:
                    logging.info(
                        f"Port {change['port']}: "
                        f"{'Closed→Open' if change['final_state'] else 'Open→Closed'}"
                    )
            else:
                logging.warning("No port state changes detected")
                
        return True

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Port Knocking Sequence Tool")
    parser.add_argument("ip", help="Target IP address")
    parser.add_argument("ports", help="Comma-separated list of ports in sequence")
    parser.add_argument("--delay", type=float, default=0.5,
                      help="Delay between knocks in seconds")
    parser.add_argument("--protocol", choices=['tcp', 'udp'], default='tcp',
                      help="Protocol to use (tcp or udp)")
    parser.add_argument("--verify", type=int,
                      help="Port to verify after knock sequence")
    args = parser.parse_args()
    
    # Parse port sequence
    port_sequence = [int(p) for p in args.ports.split(',')]
    
    knocker = PortKnocker(args.ip, port_sequence, args.delay, args.protocol)
    
    if args.verify:
        knocker.perform_advanced_knock([args.verify])
    else:
        knocker.perform_advanced_knock()
